import { Request, Response } from "express";
import logger from "../logger/logger";
import Store from "../models/store";
import { decodeBase64, getUserSession, matchPassword } from "../utils/utils";
import User from "../models/user";
import {
  isCalendarMaxFutureBookingDaysValid,
  isCalendarMinEarliestBookingTimeValid,
  isPasswordValid,
} from "../validator/validator";
import {
  isTerminPlanerGoogleCalendarConnected,
  terminPlanerRequest,
  terminPlanerRequestChangeFutureBookingDays,
} from "../utils/terminPlaner";

// this request is needed to get the store id after the user has connected his calendar to redirect him back to the calendar page
export async function GetStoreId(req: Request, res: Response) {
  try {
    const userSession = await getUserSession(req);

    if (!userSession) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // check if user has a store

    const store = await Store.findOne({
      where: {
        owner_user_id: userSession.user_id,
      },
      attributes: ["store_id"],
    });

    if (!store) {
      return res.status(401).send({ err: "unauthorized" });
    }

    res.status(200).send({ storeId: store.store_id });
  } catch (error) {
    logger.error(error);
    res.status(500).send({ err: "invalid request" });
  }
}

enum CalendarStatus {
  NOT_CONNECTED = "NOT_CONNECTED", // this is a custom status
  // the following statuses are from termin planner
  PENDING = "PENDING",
  OK = "OK",
  NOPERM = "NOPERM",
  ERROR = "ERROR",
}

export async function GetCalendarSettings(req: Request, res: Response) {
  try {
    const userSession = await getUserSession(req);

    if (!userSession) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // check if the user has the google calendar connected

    const googleCalendarConnected = await isTerminPlanerGoogleCalendarConnected(
      userSession.user_id,
      ["status"]
    );

    if (!googleCalendarConnected) {
      return res.status(200).send({
        status: CalendarStatus.NOT_CONNECTED,
      });
    }

    if (googleCalendarConnected.status !== CalendarStatus.OK) {
      return res.status(200).send({
        status: googleCalendarConnected.status,
      });
    }

    // get user

    const user = await User.findOne({
      where: {
        user_id: userSession.user_id,
      },
      attributes: [
        "store_id", // needed if the user is not a store owner to get the max and min booking days from the store
        "calendar_max_future_booking_days",
        "calendar_min_earliest_booking_time",
        "calendar_using_primary_calendar",
        "google_account_name",
        "google_account_picture",
      ],
    });

    if (!user) {
      return res.status(401).send({ err: "unauthorized" });
    }

    let userSettings = {
      calendar_using_primary_calendar: user.calendar_using_primary_calendar,
      calendar_max_future_booking_days: user.calendar_max_future_booking_days,
      calendar_min_earliest_booking_time:
        user.calendar_min_earliest_booking_time,
      google_account_name: user.google_account_name,
      google_account_picture: user.google_account_picture,
    };

    const store = await Store.findOne({
      where: {
        store_id: user.store_id,
      },
      attributes: [
        "owner_user_id", // needed to check if the user is the store owner
        "calendar_max_future_booking_days",
        "calendar_min_earliest_booking_time",
      ],
    });

    if (!store) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // if the user has not set the calendar settings yet, set them to the values from the store
    if (
      user.calendar_max_future_booking_days === null ||
      user.calendar_min_earliest_booking_time === null
    ) {
      userSettings = {
        ...userSettings,
        calendar_max_future_booking_days:
          store.calendar_max_future_booking_days,
        calendar_min_earliest_booking_time:
          store.calendar_min_earliest_booking_time,
      };
    }

    // check if user has a store

    if (store.owner_user_id !== userSession.user_id) {
      // user is not store owner - means he is an employee
      res.status(200).send({
        status: googleCalendarConnected.status,
        userSettings: userSettings,
      });
      return;
    }
    // user is store owner
    res.status(200).send({
      status: googleCalendarConnected.status,
      storeSettings: {
        calendar_max_future_booking_days:
          store.calendar_max_future_booking_days,
        calendar_min_earliest_booking_time:
          store.calendar_min_earliest_booking_time,
      },
      userSettings: userSettings,
    });
  } catch (error) {
    logger.error(error);
    res.status(500).send({ err: "invalid request" });
  }
}

export async function UpdatePersonalCalendarSettings(
  req: Request,
  res: Response
) {
  try {
    const {
      calendarUsingPrimaryCalendar,
      calendarMaxFutureBookingDays,
      calendarMinEarliestBookingTime,
    } = req.body;

    if (
      calendarUsingPrimaryCalendar === undefined ||
      calendarMaxFutureBookingDays === undefined ||
      calendarMinEarliestBookingTime === undefined ||
      !isCalendarMinEarliestBookingTimeValid(calendarMinEarliestBookingTime)
    ) {
      return res.status(400).send({ err: "invalid request" });
    }

    const userSession = await getUserSession(req);

    if (!userSession) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // check if the user has the google calendar connected

    const googleCalendarConnected = await isTerminPlanerGoogleCalendarConnected(
      userSession.user_id
    );

    if (!googleCalendarConnected) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // request to termin planner to update the settings

    const user = await User.findOne({
      where: {
        user_id: userSession.user_id,
      },
      attributes: ["store_id", "payment_plan"],
    });

    if (!user) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // check if the user has not exceeded the max future booking days

    if (
      !isCalendarMaxFutureBookingDaysValid(
        calendarMaxFutureBookingDays,
        user.payment_plan
      )
    ) {
      return res.status(400).send({ err: "invalid request" });
    }

    await terminPlanerRequest("/api/v1/changedSettings", "POST", {
      storeId: user.store_id,
    });

    // update user

    await User.update(
      {
        calendar_using_primary_calendar: calendarUsingPrimaryCalendar,
        calendar_max_future_booking_days: calendarMaxFutureBookingDays,
        calendar_min_earliest_booking_time: calendarMinEarliestBookingTime,
      },
      {
        where: {
          user_id: userSession.user_id,
        },
      }
    );

    res.status(200).send({ msg: "calendar settings updated" });
  } catch (error) {
    logger.error(error);
    res.status(500).send({ err: "invalid request" });
  }
}

export async function UpdateStoreCalendarSettings(req: Request, res: Response) {
  try {
    const { calendarMaxFutureBookingDays, calendarMinEarliestBookingTime } =
      req.body;

    // validate request

    if (
      (calendarMaxFutureBookingDays === undefined &&
        calendarMinEarliestBookingTime === undefined) ||
      (calendarMinEarliestBookingTime !== undefined &&
        !isCalendarMinEarliestBookingTimeValid(calendarMinEarliestBookingTime))
    ) {
      return res.status(400).send({ err: "invalid request" });
    }

    const userSession = await getUserSession(req);

    if (!userSession) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // get payment plan of the user

    const user = await User.findOne({
      where: {
        user_id: userSession.user_id,
      },
      attributes: ["payment_plan"],
    });

    if (!user) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // check if the user has not exceeded the max future booking days

    if (
      !isCalendarMaxFutureBookingDaysValid(
        calendarMaxFutureBookingDays,
        user.payment_plan
      )
    ) {
      return res.status(400).send({ err: "invalid request" });
    }

    // check if the user has the google calendar connected

    const googleCalendarConnected = await isTerminPlanerGoogleCalendarConnected(
      userSession.user_id
    );

    if (!googleCalendarConnected) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // verify if requester is a store master

    const store = await Store.findOne({
      where: {
        owner_user_id: userSession.user_id,
      },
    });

    if (!store) {
      logger.debug("Requester is not a store master");
      return res.status(401).send({ err: "unauthorized" });
    }

    // update store

    let update: any = {};

    if (calendarMaxFutureBookingDays) {
      update = {
        ...update,
        calendar_max_future_booking_days: calendarMaxFutureBookingDays,
      };
    }

    if (calendarMinEarliestBookingTime) {
      update = {
        ...update,
        calendar_min_earliest_booking_time: calendarMinEarliestBookingTime,
      };

      await terminPlanerRequestChangeFutureBookingDays(store.store_id);
    }

    if (Object.keys(update).length === 0) {
      return res.status(400).send({ err: "invalid request" });
    }

    await Store.update(update, {
      where: {
        store_id: store.store_id,
      },
    });

    res.status(200).send({ msg: "success" });
  } catch (error) {
    logger.error(error);
    res.status(500).send({ err: "invalid request" });
  }
}

export async function UnlinkGoogleCalendar(req: Request, res: Response) {
  try {
    const { password } = req.body;

    // validate request

    if (!password) {
      return res.status(400).send({ err: "invalid request" });
    }

    const userSession = await getUserSession(req);

    if (!userSession) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // check if the user has the google calendar connected

    const googleCalendarConnected = await isTerminPlanerGoogleCalendarConnected(
      userSession.user_id
    );

    if (!googleCalendarConnected) {
      return res.status(401).send({ err: "unauthorized" });
    }

    // decode password

    const decodedPassword = decodeBase64(password);

    if (!isPasswordValid(decodedPassword)) {
      logger.debug("Password is not valid");
      return res.status(400).send({ err: "invalid request" });
    }

    // get user to compare password

    const user = await User.findOne({
      where: {
        user_id: userSession.user_id,
      },
      attributes: ["password"],
    });

    if (!user) {
      logger.debug("User not found");
      return res.status(400).send({ err: "invalid request" });
    }

    // compare password

    const match = await matchPassword(decodedPassword, user.password);

    if (!match) {
      logger.debug("Password is not valid");
      return res.status(400).send({ err: "invalid request" });
    }

    // request to termin planner to remove google account

    await terminPlanerRequest("/api/v1/removeGoogleAccount", "POST", {
      userId: userSession.user_id,
    });

    // set user google account data to null

    await User.update(
      {
        google_account_name: "",
        google_account_picture: "",
      },
      {
        where: {
          user_id: userSession.user_id,
        },
      }
    );

    res.status(200).send({ msg: "success" });
  } catch (error) {
    logger.error(error);
    res.status(500).send({ err: "invalid request" });
  }
}
